! Parallel algorithm extension

(* Public APIs *)
(* Available parallel methods *)
$ParallelMethods := $ParallelMethodAlist[[,1]];

(* Default parallel algorithm API *)
$ParallelMethod := $ParallelMethodDefaultID;
($ParallelMethod = method_String) ^:= With[{
  sel = Select[$ParallelMethodAlist, First[#] === method&, 1]},
  If[sel <=> {},
    Unprotect[$ParallelMethodDefaultID, $ParallelMethodDefaultClass];
    $ParallelMethodDefaultID    = sel[[1,1]];
    $ParallelMethodDefaultClass = sel[[1,2]];
    Protect[$ParallelMethodDefaultID, $ParallelMethodDefaultClass]];
  $ParallelMethodDefaultID];
($ParallelMethod =.) ^:= If[Length[$ParallelMethodAlist] > 0,
  Unprotect[$ParallelMethodDefaultID, $ParallelMethodDefaultClass];
  $ParallelMethodDefaultID    = $ParallelMethodAlist[[1,1]];
  $ParallelMethodDefaultClass = $ParallelMethodAlist[[1,2]];
  Protect[$ParallelMethodDefaultID, $ParallelMethodDefaultClass]];

(* Algorithm assessor *)
Member[Parallel, member_] ^:= (
  If[Not[StringQ[$ParallelMethod]], $ParallelMethod =.];
  Member[$ParallelMethodDefaultClass, member]);

Parallel[algorithm_String] := With[{
  sel = Select[$ParallelMethodAlist, First[#] === algorithm&, 1]},
  If[sel <=> {}, sel[[1,2]], Parallel]];

(* Detail implementation *)
(* Helper class for parallel algorithm writter *)
Algorithm$ParallelHelper = Class[{},
  { DefaultSharedSize = 256
    }, {},

  (* Attribute initializer *)
  InitializeAttributes[] := (
!    SetAttributes[{Map, MapThread, MapIndexed, MapAll, MapAt}, {HoldFirst}];
    );

  (* Number of Processor *)
  NumberOfProcessors[] := With[{
    ncpu = Check[Switch[Literal[Map][BuildInfo,
      {"Target:OS_TYPE", "Target:OS_NAME"}],
      (* Sun Solaris *)
      {"SunOS",		_},	Get["!/usr/sbin/psrinfo -p"],

      (* 4.4BSD-Lite family *)
      {_,	"Darwin"},	Get["!/usr/sbin/sysctl -n hw.ncpu"],
      {"4.4BSD-Lite",	_},	Get["!/sbin/sysctl -n hw.ncpu"],

      (* GNU/Linux *)
      {"Linux",		_},	Get["!/bin/cat /proc/cpuinfo | /bin/grep ^processor | /usr/bin/wc -l"],

      (* Otherwise *)
      _,			Null], Null]},
    If[RealQ[ncpu], ncpu, Null]];

  (* Split list `l' into `n' sub-lists *)
  Split[l_, n_Real] := Module[{N = Length[l], m, r, i, Map = Literal[Map]},
    m = Floor[N / n]; r = N - m * n; i = 0;
    Map[With[{i0 = i}, i = i0 + #; Take[l, {i0 + 1, i}]]&, Table[m, {n}] +
      Take[Join[Table[1, {n - 1}], Table[0, {n}]], {n - r, n + n - r - 1}]]];

  (* Wait4 wrapper *)
  Wait4[arg___] := With[{result = Literal[Wait4][arg]},
    Append[Take[result, 2], Drop[result, 2]]];

  (* Default algorithm template *)
  Map[f_, l_, opt:((_Rule|_RuleDelayed)...)] := Literal[Map][f, l];
  Map[f_, l_, level:_Real, opt:((_Rule|_RuleDelayed)...)] := Literal[Map][f, l, level];
  Map[f_, l_, level:{_Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[Map][f, l, level];
  Map[f_, l_, level:{_Real, _Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[Map][f, l, level];

  MapThread[f_, l_, opt:((_Rule|_RuleDelayed)...)] := Literal[MapThread][f, l];
  MapThread[f_, l_, level:_Real, opt:((_Rule|_RuleDelayed)...)] := Literal[MapThread][f, l, level];
  MapThread[f_, l_, level:{_Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[MapThread][f, l, level];
  MapThread[f_, l_, level:{_Real, _Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[MapThread][f, l, level];

  MapIndexed[f_, l_, opt:((_Rule|_RuleDelayed)...)] := Literal[MapIndexed][f, l];
  MapIndexed[f_, l_, level:_Real, opt:((_Rule|_RuleDelayed)...)] := Literal[MapIndexed][f, l, level];
  MapIndexed[f_, l_, level:{_Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[MapIndexed][f, l, level];
  MapIndexed[f_, l_, level:{_Real, _Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[MapIndexed][f, l, level];

  MapAll[f_, l_, opt:((_Rule|_RuleDelayed)...)] := Literal[MapAll][f, l];

  MapAt[f_, l_, n:_Real, opt:((_Rule|_RuleDelayed)...)] := Literal[MapAt][f, l, n];
  MapAt[f_, l_, n:{__Real}, opt:((_Rule|_RuleDelayed)...)] := Literal[MapAt][f, l, n];
  MapAt[f_, l_, n:{{__Real}..}, opt:((_Rule|_RuleDelayed)...)] := Literal[MapAt][f, l, n];

  ];

(* Setup algorithm class autoload *)
Library@AutoLoad[Algorithm$Parallel$Fork, "Algorithm/Parallel/Fork"];

(* Associative list of parallel algorithm *)
$ParallelMethodAlist = {
  "Fork":>Algorithm$Parallel$Fork,
  Null[]};
Protect[$ParallelMethodAlist];

! End of File
