$ElementValues = Class[{},
  {
    },

  {
    (* CouplingMatrix Cache *)
    RefreshDerivatives,
    PreviousFunctions,
    Derivatives,
    EvalCouplingMatrix,
    PreviousVariables,
    PreviousVariableIndexes,
    PreviousEvalCouplingMatrix,

    (* ElementValues Cache *)
    RefreshElementValues,
    HoldElementValues,
    EvalSetCoupledElements,
    $CoupledKeys,
    $VariableRules,
    Keys,
    Functions,
    Variables,
    Debugging = False
    },

  Constructor[] := (
    SetAttributes[SetDelayedElementValues, HoldAll];
    InvalidateCachedElementValues[];
    InvalidateCachedDerivatives[];
    );

  (* For debugging *)
  Debug[argv___] := If[Debugging, Literal[Print]["DEBUG: ", argv]];

  Dump[] := (
    Print[">>>> Begin ElementVlaues Dump"];
    Print[{
      RefreshDerivatives,
      PreviousFunctions,
      Derivatives,
      EvalCouplingMatrix,
      PreviousVariables,
      PreviousVariableIndexes,
      PreviousEvalCouplingMatrix,

      RefreshElementValues,
      HoldElementValues,
      EvalSetCoupledElements,
      $CoupledKeys,
      $VariableRules,
      Keys,
      Functions,
      Variables}];
    Print[">>>> End   ElementVlaues Dump"];
    );

  (* Public API *)
  SetCoupledElements[] := (Compile[]; Debug["SetCoupledElements"]; EvalSetCoupledElements[]);

  CoupledKeys[]        := (Compile[]; Debug["CoupledKeys"]; $CoupledKeys);

  CouplingMatrix[]     := (Compile[]; Debug["CouplingMatrix"]; If[Functions === {}, Return[{False, {}}],
    With[{current = CurrentVariableIndexes[]}, Which[
      current === {},		{False, Table[{}, {Length[Functions]}]},

      Max[current] < 0,		{False, Table[0, {Length[Functions]}, {Length[current]}]},

      current === PreviousVariableIndexes,	(
	With[{val = PreviousEvalCouplingMatrix[]},
	  {Position[val, _?(#<>0&), {2}, 1] <=> {}, val}]),

      CompileDerivatives[];
      Derivatives === Null,	{False, Table[{}, {Length[Functions]}]},

      True,	With[{expr = Transpose[Part[Derivatives, current]]},
	PreviousVariableIndexes = current;
	PreviousEvalCouplingMatrix = (expr)&/.{Hold->Null};

	With[{val = PreviousEvalCouplingMatrix[]},
	  {Position[val, _?(#<>0&), {2}, 1] <=> {}, val}]]]]]);

  (* Setter/Getter proxy for ElementValues *)
  GetElementValues[] := If[
Debug["GetElementValues"];

    HoldElementValues === Null,
    Unevaluated$[Literal[ElementValues]],
    First[HoldElementValues]];

  SetElementValues[l_] := (InvalidateCachedElementValues[];
Debug["SetElementValues"];
    HoldElementValues = Hold[l]; Unevaluated$[l]);

  SetDelayedElementValues[l_] := (InvalidateCachedElementValues[];
Debug["SetDelayedElementValues"];
    HoldElementValues = Hold[l]; Unevaluated$[l]);

  UnsetElementValues[] := (InvalidateCachedElementValues[];
Debug["UnsetElementValues"];
    HoldElementValues = Null);

  DefinitionElementValues[] := If[
Debug["DefinitionElementValues"];

    HoldElementValues === Null,
    Hold[{ElementValues := ElementValues}],
    With[{$HoldElementValues = HoldElementValues},
      ReplacePart[Hold[{ElementValues := $HoldElementValues}], Null, {1, 1, 2, 0}]]];

  (* CouplingMatrix Cache *)
  InvalidateCachedDerivatives[] := (
    PreviousVariables = Null;
    PreviousVariableIndexes = Null;
    RefreshDerivatives = True);

  CompileDerivatives[] := Which[
    Not[RefreshDerivatives],		(* Nothing TODO *),

    Functions === PreviousFunctions,	RefreshDerivatives = False,

    True,				(
      Derivatives = Which[
	Functions === {},	Null,

	Variables === {},	Null,

	True,			Append[
	  Thread[Apply[With[{func = #1, masks = #2}, Switch[func,

	    _Hold,	With[{expr = Extract[func, {1}, Unevaluated$],
	      replaces = First[SwitchCases[$VariableRules,
		{RuleDelayed[Alternatives@@MapThread[If[#1, Identity[#2], Null[]]&, {masks, Variables}], _]}]]},
	      MapThread[If[Not[#1], 0, With[{var = #2},
		Hold[D[expr, var]/.replaces]]]&, {masks, Variables}]],

	    _,		(
	      MapThread[If[Not[#1], 0, With[{d = D[func, #2]},
		Hold[d]/.$VariableRules]]&, {masks, Variables}])]]&, Functions, {1}]],
	  Table[0, {Length[Functions]}]]];
      PreviousFunctions = Functions;
      RefreshDerivatives = False)];

  CurrentVariables[] := Apply[#2[#1]&, FFS["VAR"][[,{1, 2}]], {1}];

  CurrentVariableIndexes[] := With[{current = CurrentVariables[]},
    Which[
      current === {},			{},

      And[PreviousVariables <=> Null,
	current === First[PreviousVariables]],	Second[PreviousVariables],

      True,				(
	PreviousVariables = {current, With[{pos = Position[Variables, #, 1, 1]},
	  If[pos === {}, -1, pos[[1,1]]]]&/@current};
	Second[PreviousVariables])]];

  (* ElementValues Cache *)
  InvalidateCachedElementValues[] := (
    RefreshElementValues = True);

  Compile[] := Which[
    Not[RefreshElementValues],		(* Nothing TODO *),

    HoldElementValues === Null,		(
      $CoupledKeys = $VariableRules = {};
      Keys = Functions = Variables = {};
      EvalSetCoupledElements = (Null&);
      InvalidateCachedDerivatives[];
      RefreshElementValues = False),

    True,	Module[{db},
      (* Build internal database from HoldElementValues *)
      db =Map[Switch[#,
	(Rule|RuleDelayed)[_String[_String], _],	With[{
	  key = #[[1, 0]], element = #[[1, 1]],
	  pos = With[{elm = #[[1, 1]]}, Flatten[{
	    If[StringPosition[elm, "."] === {}, -Element["POSITION", elm], LINE["POSITION", elm]]}]],
	  lazy = SameQ[Head[#], RuleDelayed],
	  expr = If[SameQ[Head[#], RuleDelayed], Extract[#, {2}, Hold], Second[#]]},
	  {lazy, pos, key[element], expr, Union[Cases[expr, _String[_String], Infinity]]}],

	RuleDelayed[_, _],				With[{
	  expr = Extract[#, {2}, Hold]},
	  {True,  {},      Command, expr, Union[Cases[expr, _String[_String], Infinity]]}],

	_,	Null[]]&, First[HoldElementValues]];

      (* Reduce lazy expression into early form if possible[DISABLED:too expensive] *)
      (* ASSUMPTION: expr MIGHT NOT change environment *)
If[False,
      db = Apply[With[{lazy = #1, pos = #2, key = #3,
	expr = #4, reduce = If[#1 && #2 <=> {}, ReleaseHold[#4], Null], vars = #5}, Which[
	  Not[lazy] || pos === {},	{lazy,  pos, key, expr,   vars},
	  expr === Hold[reduce],	{False, pos, key, reduce, vars},
	  True,				{lazy,  pos, key, expr,   vars}]]&, db, {1}];
  ];

      With[{rules = SwitchCases[Map[With[{key = Head[#], element = First[#],
	pos = With[{elm = First[#]}, Flatten[{
	  If[StringPosition[elm, "."] === {}, -Element["POSITION", elm], LINE["POSITION", elm]]}]]},
	With[{p1 = If[pos === {}, 0, First[pos]]}, Which[
	  p1 > 0,	key[element]:>LINE[   key,  p1],
	  p1 < 0,	key[element]:>Element[key, -p1],
	  True,	key[element]->0]]]&, Union[Flatten[db[[, 5]]]]], {_Rule, _}]},
	(* Zeoring non-existent element variables in database *)
	db = Apply[With[{lazy = #1, pos = #2, key = #3, expr = #4, vars = DeleteCases[#5, 0]},
	  {lazy, pos, key, expr, vars}]&,
	  db/.First[rules], {1}];

	(* Build variable replace RuleDelayed *)
	$VariableRules = Last[rules]];

      (* Extract existing Keys *)
      With[{hit = Select[db, #[[2]] <=> {}&]}, Module[{hash},
	$CoupledKeys = Apply[{Head[#3], #2}&, hit, {1}];
	Keys      = hit[[, 3]];
	Variables = Union[Flatten[hit[[, 5]]]];
	Functions = Thread[{hit[[, 4]], Map[With[{vars = #},
	  Clear[hash]; hash[_] := False;
	  Scan[(hash[#] = True)&, vars];
	  Map[hash, Variables]]&, hit[[, 5]]]}];
	]];

      (* Build SetCoupedElements function body from database *)
      EvalSetCoupledElements = With[{exprs = CompoundExpression@@Apply[With[{
	p1 = If[#2 === {}, 0, First[#2]], key = Head[#3],
	(* WORKAROUND for General::invset at With binding evaluation[FIXME] *)
	(* Wrap 1st argument of If[#1, ...] by With binding *)
	expr = With[{flag = #1}, If[flag, Extract[#4, {1}, Unevaluated$], #4]]}, Which[
	  p1 > 0,	Unevaluated$[LINE[   key,  p1] = expr],
	  p1 < 0,	Unevaluated$[Element[key, -p1] = expr],
	  True,		Unevaluated$[expr]]]&, db, {1}]},
	(exprs)&/.$VariableRules];

      (* Update cache state *)
      InvalidateCachedDerivatives[];
      RefreshElementValues = False]];

  ];

(* Create instance for ElementValues management *)
The$ElementValues = $ElementValues[];

(* Invalidate cache by FFS/BeamLineChanged Package event *)
PkgHookAdd["FFS/BeamLineChanged", The$ElementValues@InvalidateCachedElementValues[]];

(* Import initial value from ElementValues *)
The$ElementValues@SetElementValues[ElementValues];

(* Override ElementValues *)
Unprotect[ElementValues];
Clear[ElementValues];
ElementValues             := The$ElementValues@GetElementValues[];
(ElementValues   =.    ) ^:= The$ElementValues@UnsetElementValues[];
(ElementValues   = var_) ^:= The$ElementValues@SetElementValues[var];
(ElementValues  := var_) ^:= The$ElementValues@SetDelayedElementValues[var];
(?ElementValues        ) ^:= The$ElementValues@DefinitionElementValues[];
(Clear[a___, ElementValues, b___]) ^:= (ElementValues =.; Clear[a, b]);
Protect[ElementValues];

(* Define proxy for SetCoupledElements, CouplingMatrix, CoupledKeys *)
SetCoupledElements := The$ElementValues@SetCoupledElements[];
CouplingMatrix     := The$ElementValues@CouplingMatrix[];
CoupledKeys        := The$ElementValues@CoupledKeys[];
Protect[SetCoupledElements, CouplingMatrix, CoupledKeys];

! End of File
