(* FFS DRAW/GEO/TDR command backend *)
FFS$DRAW[] := FFS$DRAW[$DisplayFunction];
FFS$GEO[] := FFS$GEO[$DisplayFunction];
FFS$TDR[] := FFS$TDR[$DisplayFunction];

(* Default fallback *)
FFS$DRAW[drawer_] := With[{next = First[FFS$DRAW[Command->OpticsPlot]]},
  Print["DRAW command via ", drawer, " is not supported"]; next];
FFS$GEO[drawer_] := (
  Print["GEO command via ", drawer, " is not supported"]; "");
FFS$TDR[drawer_] := (
  Print["TDR command via ", drawer, " is not supported"]; "");

(* Search beamline position by given pattern *)
FFS$DRAW["POSITION", pat_String] := If[pat == "", Null, With[{p1 = LINE["POSITION", pat]}, Switch[p1,
  _Real,	p1,
  {__Real},	First[p1],
  _,		With[{p2 = LINE["POSITION", ToUpperCase[pat]]}, Switch[p2,
    _Real,	p2,
    {__Real},	First[p2],
    _,		Null]]]]];

(* Parse DRAW command line: [begin [end]] optical-function-list *)
FFS$DRAW[Command->Unevaluated$[OpticsPlot]] := Module[{r1, r2, w, w1, list},
  Which[
    Not[RealQ[r1 = FFS$DRAW["POSITION",
      (w = Read[-1, Word, ReadNewRecord->False])]]],	(	(* No region part *)
	r1 = 1;
	r2 = LINE["LENGTH"]),

    Not[RealQ[r2 = FFS$DRAW["POSITION",
      (w = Read[-1, Word, ReadNewRecord->False])]]],	(	(* No region end *)
	r2 = LINE["LENGTH"]),

    True,	(w = Read[-1, Word, ReadNewRecord->False])];	(* Read next word after region pair *)

  w1 = "&";
  list = Table[If[w1 <> "&", Break[], Table[(
    {w1, w} = With[{p = Min[StringPosition[w, "&", 1], StringPosition[w, ";", 1]]}, Switch[p,
      Infinity,	{w, ""},
      1,	{w[1], w[2, -1]},
      _,	{w[1, p-1], w[p, -1]}]];
    Which[
      w1 == "" && w == "" && (w = Read[-1, Word, ReadNewRecord->False]) == "",	(w1 = ";"; Break[]),
      w1 == "",		Null[],
      w1 == "&",	Break[],
      w1 == ";",	Break[],
      Twiss$Label[ToUpperCase[w1]] <=> "",	ToUpperCase[w1],
      True,					(w = w1; Break[])]), {Infinity}]], {Infinity}];

  list = { list,
    Region->{r1, r2},
    Which[
      w1 == ";",						(w = ""; Lattice->False),
      w == "",							(w = ""; Lattice->False),
      LINE["POSITION", w] <=> {},				With[{pat = w},
	w = ""; Null[Lattice->True, Names->pat]],
      LINE["POSITION", ToUpperCase[w]] <=> {},			With[{pat = ToUpperCase[w]},
	w = ""; Null[Lattice->True, Names->pat]],
      ToUpperCase[w] == "LAT" || ToUpperCase[w] == "LATTICE",	(w = ""; Lattice->True),
      True,							Lattice->False],
    GridLines->{None,Automatic}, PlotLabel:>TITLE};

  w->list];

Protect[FFS$DRAW, FFS$GEO, FFS$TDR];

! End of File
