(* PkgConfigMatch[] public API *)
PkgConfigMatch[pkg_String, key_String] := Select[
  Union[PkgConfig$UserDB[pkg][[,1]], PkgConfig$SystemDB[pkg][[,1]]],
  StringMatchQ[#, key]&];

(* PkgConfig[]/PkgConfigList[] public API *)
PkgConfigList[l__] := Scan[With[{pkg = #[[1]], config = #[[2]]},
  Scan[With[{var = #[[1]], val = #[[2]]},
    PkgConfig[pkg, var] = val]&,
    First[SwitchCases[config, {Rule[_String, _]}]]]]&,
  First[SwitchCases[Flatten[{l}], {Rule[_String, _List]}]]];
PkgConfig[pkg_String, vl:{___String}] := With[{pl = Map[Rule[#, PkgConfig$Type[pkg, #]]&, vl]},
  MapThread[Which[
    #1 <=> {},	#1[[1, 2]],
    #2 <=> {},	#2[[1, 2]],
    True,	Null]&,
    { SwitchCases[PkgConfig$UserDB[  pkg], pl],
      SwitchCases[PkgConfig$SystemDB[pkg], pl]}]];
PkgConfig[pkg_String, var_String] := With[{pattern = Rule[var, PkgConfig$Type[pkg, var]]},
  Apply[Which[
    #1 <=> {},	#1[[1, 2]],
    #2 <=> {},	#2[[1, 2]],
    True,	Null]&, {
      First[SwitchCases[PkgConfig$UserDB[  pkg], {pattern}]],
      First[SwitchCases[PkgConfig$SystemDB[pkg], {pattern}]]}]];
(PkgConfig[pkg_String, var_String] =.) ^:= Module[{drop, rem},
  {drop, rem} = SwitchCases[PkgConfig$UserDB[pkg],
    {Rule[var, _], _}];
  PkgConfig$UserDB[pkg] = rem];
(PkgConfig[pkg_String, var_String] = val_) ^:= With[{
  types = PkgConfig$Type[pkg, var]},
  Switch[val,
    types, Module[{drop, rem},
      {drop, rem} = SwitchCases[PkgConfig$UserDB[pkg],
	{Rule[var, _], _}];
      PkgConfig$UserDB[pkg] = AppendTo[rem, Rule[var, val]]],
    _, PkgConfig$UserDB[pkg]]];

(* with prefix wrapper *)
PkgConfig[pkg_String, prefix_String, vl:{___String}] := PkgConfig[pkg, Map[prefix//#&, vl]];
PkgConfig[pkg_String, prefix_String, var_String] := PkgConfig[pkg, prefix//var];
(PkgConfig[pkg_String, prefix_String, var_String] =.) ^:= (PkgConfig[pkg, prefix//var] =.);
(PkgConfig[pkg_String, prefix_String, var_String] = val_) ^:= (PkgConfig[pkg, prefix//var] = val);

(* PkgConfig$Type[]/PkgConfig$TypeList[] package API *)
PkgConfig$TypeList[l__] := Scan[With[{pkg = #[[1]], config = #[[2]]},
  Scan[With[{var = #[[1]], type = #[[2]]},
    PkgConfig$Type[pkg, var] = type]&,
    First[SwitchCases[config, {Rule[_String, _]}]]]]&,
  First[SwitchCases[Flatten[{l}], {Rule[_String, _List]}]]];
PkgConfig$Type[pkg_String, var_String] := With[{
  types = PkgConfig$TypeDB[pkg, var]},
  If[types === Null, _, types]];
(PkgConfig$Type[pkg_String, var_String] =.) ^:= (
  PkgConfig$TypeDB[pkg, var] =.);
(PkgConfig$Type[pkg_String, var_String] = val_) ^:= (
  PkgConfig$TypeDB[pkg, var] = val);

(* PkgConfig$Default[]/PkgConfig$DefaultList[] package API *)
PkgConfig$DefaultList[l__] := Scan[With[{pkg = #[[1]], config = #[[2]]},
  Scan[With[{var = #[[1]], val = #[[2]]},
      PkgConfig$Default[pkg, var] = val]&,
    First[SwitchCases[config, {Rule[_String, _]}]]]]&,
  First[SwitchCases[Flatten[{l}], {Rule[_String, _List]}]]];
PkgConfig$Default[pkg_String, vl:{___String}] := Map[If[# <=> {}, #[[1, 2]], Null]&,
  SwitchCases[PkgConfig$SystemDB[pkg],
    Map[Rule[#, PkgConfig$Type[pkg, #]]&, vl]]];
PkgConfig$Default[pkg_String, var_String] := With[{hit = First[
  SwitchCases[PkgConfig$SystemDB[pkg], {Rule[var, PkgConfig$Type[pkg, var]]}]]},
  If[hit <=> {}, hit[[1,2]], Null]];
(PkgConfig$Default[pkg_String, var_String] =.) ^:= Module[{drop, rem},
  {drop, rem} = SwitchCases[PkgConfig$SystemDB[pkg],
    {Rule[var, _], _}];
  PkgConfig$SystemDB[pkg] = rem];
(PkgConfig$Default[pkg_String, var_String] = val_) ^:= With[{
  types = PkgConfig$Type[pkg, var]},
  If[PkgConfig$TypeDB[pkg, var] === Null,
    (* Automatic type definition *)
    PkgConfig$Type[pkg, var] = Switch[val,
      _Real,		_Real,
      _String,		_String,
      {__Real},		{___Real},
      {__String},	{___String},
      _List,		_List,
      _,		_]];
  Switch[val,
    types, Module[{drop, rem},
      {drop, rem} = SwitchCases[PkgConfig$SystemDB[pkg],
	{Rule[var, _], _}];
      PkgConfig$SystemDB[pkg] = AppendTo[rem, Rule[var, val]]],
    _, PkgConfig$SystemDB[pkg]]];

(* with prefix wrapper *)
PkgConfig$Default[pkg_String, prefix_String, vl:{___String}] := PkgConfig$Default[pkg, Map[prefix//#&, vl]];
PkgConfig$Default[pkg_String, prefix_String, var_String] := PkgConfig$Default[pkg, prefix//var];
(PkgConfig$Default[pkg_String, prefix_String, var_String] =.) ^:= (PkgConfig$Default[pkg, prefix//var] =.);
(PkgConfig$Default[pkg_String, prefix_String, var_String] = val_) ^:= (PkgConfig$Default[pkg, prefix//var] = val);

(* PkgHook package hook API *)
PkgHookAdd[pkg_String, hooks__] := PkgHookAppend[pkg, hooks];
PkgHookPrepend[pkg_String, hooks__] := Switch[PkgConfig$HookDB[pkg],
  Automatic,	Reverse[Unevaluated$[CompoundExpression[Null, hooks]]],
  _List,	(PkgConfig$HookDB[pkg] = Join[
    Reverse[MapThread[RuleDelayed, {Command, Unevaluated$[{hooks}]}]], PkgConfig$HookDB[pkg]]),
  _,		$Failed];
PkgHookAppend[pkg_String, hooks__] := Switch[PkgConfig$HookDB[pkg],
  Automatic,	CompoundExpression[hooks, Null],
  _List,	(PkgConfig$HookDB[pkg] = Join[
    PkgConfig$HookDB[pkg], MapThread[RuleDelayed, {Command, Unevaluated$[{hooks}]}]]),
  _,		$Failed];
PkgHookEval[pkg_String] := Scan[Second, PkgConfig$HookDB[pkg]];
PkgHookEvalClear[pkg_String] := (PkgHookEval[pkg]; PkgConfig$HookDB[pkg] =.);
PkgHookEvalAfter[pkg_String] := (PkgHookEval[pkg]; PkgConfig$HookDB[pkg] = Automatic);

SetAttributes[{PkgHookAdd, PkgHookPrepend, PkgHookAppend}, HoldRest];

(* PkgConfig internal database *)
PkgConfig$UserDB[_]    := {};
PkgConfig$SystemDB[_]  := {};
PkgConfig$TypeDB[_, _] := Null;
PkgConfig$HookDB[_]    := {};

! End of File
