`Simplex;
SetAttributes[{`MinimumSize,`ExpansionRatio,`TryReflect,`Persistence,`QuickShrink,`ExtendLevel},Constant];

BeginPackage[Optimize`,Global`]

Begin[Optimize`];

Simplex=Class[{},{},{CurrentSimplex,Status="Stop"},

  Amoebar[p_,f_,r_,limit_,kr_]:=Module[
    {pr,fac1=(1-r)/(Length[p]-1),fr},
    pr=MapThread[Restrict,{Plus@@p[[,2]]*fac1+p[[kr,2]]*(r-fac1),limit[[1]],limit[[2]]}];
    {pr,fr=f[pr],
      If[fr<p[[kr,1]],
        Sort[ReplacePart[p,{fr,pr},kr]],
        p]}];

  Amoeba[p_,f_,limit_,per_,minq_,exp_,cont_,quick_,exl_]:=Module[
    {pr,pr1,pl,fr,fr1,prr,n,kr,fr0,pr0,pt},
    Do[
      Status=StandardForm["Reflect "//kr];
      {pt,fr,pr}=Amoebar[p,f,-1,limit,-kr];
(*      Print["dh1 ",{fr,p[[exl,1]]}]; *)
      If[kr==1,fr0=fr;pr0=pr];
      If[fr <= p[[exl,1]],
        Status="Extend";
        {pt,fr1,pr1}=Amoebar[p,f,-exp,limit,-kr];
        Return[If[fr1 <= fr,
          pr1,
          pr]]],
(*          If[fr1<=pr[[-1,1]],
            Sort[ReplacePart[pr,{fr1,pt},-1]],
            pr]]]], *)
      {kr,cont}];
    If[cont<>1,fr=fr0;pr=pr0];
    If[Length[p]>2 && fr < p[[-2,1]],
      pr,
      If[fr < p[[-1,1]],
        Status="Contract+",
        Status="Contract-"];
      {pt,fr1,pr1}=Amoebar[pr,f,0.5,limit,-1];
      If[fr1 < pr[[-1,1]],
        pr1,
        pl=pr[[1,2]];
        prr=Thread[(pl+Thread[Drop[pr,1][[,2]]])/2];
        prr=(MapThread[(If[#2<#<#3,(#>#4)*(#3-#2)+#2,#])&,
          {#,pl-minq,pl+minq,pl}]&/@prr);
        Do[
          Status=StandardForm["Shrink "//n];
          fv=f[prr[[n-1]]];
          pr[[n]]={fv,prr[[n-1]]};
          If[quick && (fv<pr[[n-1,1]] || fv>pr[[-1,1]]),
            Break[]],
          {n,Length[pr],2,-1}];
        If[per<>1,
          pr[[1]]={pr[[1,1]]*per+f[pl]*(1-per),pl}];
        Sort[pr]]]];

  DownhillSimplex[p0_,f_,opt___]:=Module[
    {def={VariableRange:>{Table[Max[],{Length[p0[[1,2]]]}],
      Table[Min[],{Length[p0[[1,2]]]}]},
      MaxIteration:>Max[100,10*Length[p0]],
      ExtendLevel->1,
      Output->6,MinimumSize->0,ExpansionRatio->2,TryReflect->1,
      Tolerance->1e-6,Persistence->1,QuickShrink->0},
      limit,n,tol,per,minq,exp,cont,quick,exl,
      lo,p=Sort[p0],nc=Floor[(PageWidth-3)/22]},
    {limit,n,lo,tol,per,minq,exp,cont,quick,exl}=
      {VariableRange,MaxIteration,Output,
        Tolerance,Persistence,MinimumSize,ExpansionRatio,TryReflect,QuickShrink,
        ExtendLevel}/.{opt}/.def;
    CurrentSimplex=p;
    exl=Restrict[exl,1,Length[p]];
    cont=Restrict[cont,1,Length[p]];
    Check[
      Catch[
        Do[CurrentSimplex=p=Amoeba[p,f,limit,per,minq,exp,cont,quick,exl];
	  Switch[lo,
	    _Real|{_Real, _Real},	StandardForm[$FORM='10.6';
              Write[lo,Drop[p,{nc+1,Length[p0]-nc}][[,1]]]]];
          If[(p[[-1,1]]-p[[1,1]])/(1e-100+Abs[p[[-1,1]]]+Abs[p[[1,1]]]) < tol,
            Break[]],{n}]],];
    Status="Stop";
    p];

  ];

End[];

EndPackage[];

Clear[The$Simplex];
SetAttributes[The$Simplex, Dynamic];

DownhillSimplex = Block[{The$Simplex},
  The$Simplex = Simplex[];
  The$Simplex@DownhillSimplex[##]]&;

SetAtributes[{DownhillSimplex,Simplex},Constant];

Default$Optimize$Function=((FFS["CALC"];1+MatchingResidual)&);

(* for backward compatibility *)
SetAttributes[{CalculateMatching, Optimize$Function, OptimizeVlist}, Dynamic];

!CalculateMatching[v_, vlist_]:=Module[{},
!  Scan[(Element[#[[1,2]], #[[1,1]]] = #[[2]])&, Thread[{vlist, v}]];
!  Optimize$Function[]];

With[{
  parallelQ = Library@Require["Algorithm/Parallel"],
  OpticsVariableList = (Select[FFS["VAR"], Or[
    #[[8]] == #[[1]],
    #[[8]] == #[[1]]//".1",
    #[[8]] == "",
    False]&]&),

  Simplex$Epsilon = 1e-12,

  def = {
    SimplexSize->0.03,
    OptimizeFunction->Default$Optimize$Function,
    InitialSimplex->Null}},

  SetupOptimize[size_, f_] := SetupOptimize[size, f, Null];

  SetupOptimize[size_, f_, p0_] := SetupOptimize[size, f, p0,
    OpticsVariableList[]];

  SetupOptimize[size_, f_, p0_, vl$_List] := With[{vl = Transpose[vl$]},
    If[p0 === Null,

      With[{ (* Generate initial simplex *)
	eps   = Simplex$Epsilon,
	limit = Take[vl, {6, 7}], (* lower, upper limit *)
	vzero = MapThread[Restrict[#3, #6, #7]&, vl]}, Module[{
	  p = Table[vzero, {Length[vzero] + 1}]},
	  ScanThread[With[{i = #1, v0 = #2, lower = #3, upper = #4},
	    p[[i, i]] = If[v0 - lower < upper - v0,
	      Min[v0 + Max[eps, size * (upper - lower), Abs[size * v0]], upper],
	      Max[v0 - Max[eps, size * (upper - lower), Abs[size * v0]], lower]];
	    ]&, {Range[Length[vzero]], vzero, Null@@limit}];
	  {Sort[With[{eval = {f[#], #}&, size = SharedSize[{-1, Range[Length[First[p]]]}]},
	    If[parallelQ, Parallel@Map[eval, p, SharedSize->size], eval/@p]]], limit}]],

      With[{ (* Use given simplex *)
	limit = Take[vl, {6, 7}]}, (* lower, upper limit *)
	{p0, limit}]]];

  OptimizeOptics[opt$:((_Rule|_RuleDelayed)...)] := Module[{
    opt = Override[opt$, def], vl, p, f},

    vl = OpticsVariableList[];

    If[Length[vl] < 1,
      Message[OptimizeOptics::novariables]; Return[$Failed]];

    With[{
      setter = Map[If[StringPosition[#, "."] === {}, Element, LINE]&, vl[[, 1]]],
      vlist = vl[[, {1, 2}]],
      limit = Transpose[vl[[, {6, 7}]]],
      size = SimplexSize/.opt, initial = InitialSimplex/.opt,
      of = OptimizeFunction/.opt},

      f = (ScanThread[(#1[#2[[2]], #2[[1]]] = #3)&, {setter, vlist, #}]; of[])&;

      (* for backward compatibility *)
      Block[{CalculateMatching, Optimize$Function, OptimizeVlist},
	CalculateMatching = (
	  ScanThread[If[StringPosition[#1[[1]], "."] === {},
	    Element[#1[[2]], #1[[1]]] = #2,
	    LINE[#1[[2]], #1[[1]]] = #2]&, {#2, #}];
	  of[])&;
	Optimize$Function = of;
	OptimizeVlist     = vlist;

	p = With[{f, vl}, DownhillSimplex[
	  Unevaluated[First[SetupOptimize[size, f, initial, vl]]],
	  f, Null@@opt, VariableRange->limit]];
	f[p[[1,2]]];
	]];

    p];

  ];

OptimizeOptics::novariables="No matching variables";

Protect[SetupOptimize,OptimizeOptics];
